import datetime as dt
import calendar
import sqlalchemy as sa
import pandas as pd


MSSQL_CONN = "mssql+pymssql://readdnds:reader%402021@121.37.138.1:14331/dnds"
MYSQL_CONN = "mysql+pymysql://reader:Reader#2020@121.37.138.1:13316/vision_product"

RISK_STYLES = ['BETA',
               'MOMENTUM',
               'SIZE',
               'EARNYILD',
               'RESVOL',
               'GROWTH',
               'BTOP',
               'LEVERAGE',
               'LIQUIDTY',
               'SIZENL']

INDUSTRY_STYLES = [
    'Bank',
    'RealEstate',
    'Health',
    'Transportation',
    'Mining',
    'NonFerMetal',
    'HouseApp',
    'LeiService',
    'MachiEquip',
    'BuildDeco',
    'CommeTrade',
    'CONMAT',
    'Auto',
    'Textile',
    'FoodBever',
    'Electronics',
    'Computer',
    'LightIndus',
    'Utilities',
    'Telecom',
    'AgriForest',
    'CHEM',
    'Media',
    'IronSteel',
    'NonBankFinan',
    'ELECEQP',
    'AERODEF',
    'Conglomerates'
]


def read_sql(query, source="mssql"):
    if source == "mssql":
        engine = sa.create_engine(MSSQL_CONN, connect_args={"charset": "GBK"})
    elif source == "mysql":
        engine = sa.create_engine(MYSQL_CONN)
    return pd.read_sql(query, con=engine)


def nearest_report_date(date):
    if len(date) == 10:
        date = date.replace("-", "")

    mm = int(date[4:6])
    year = int(date[0:4])
    last_year = year

    q = (mm // 3 + 1) * 3
    last = calendar.monthrange(last_year, q)[1]
    return dt.date(last_year, q, last).strftime("%Y%m%d")


def fd_alive_funds(trade_dt, class_code=None):
    if class_code:
        query = f"""
                select SECODE FROM TQ_FD_TYPE
                WHERE
                    ISVALID = 1 AND
                    TYPESTYLE = '{str(class_code)[0]}' AND
                    CLASSCODE = '{class_code}' AND
                    BEGINDATE <= '{trade_dt}' AND
                    (ENDDATE>='{trade_dt}' or ENDDATE = '19000101') 
                ORDER BY SECODE;
                """
    else:
        query = f"""
                select SECODE FROM TQ_FD_TYPE
                WHERE
                    ISVALID = 1 AND
                    BEGINDATE <= '{trade_dt}' AND
                    (ENDDATE>='{trade_dt}' or ENDDATE = '19000101') 
                ORDER BY SECODE;
                """
    sec_codes = read_sql(query)["SECODE"].unique().tolist()
    return sec_codes


def fd_basicinfo(security_ids=None, trade_dt=None):
    if not trade_dt:
        trade_dt = dt.datetime.today().strftime("%Y%m%d")
    else:
        trade_dt = trade_dt.replace("-", "")

    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    elif not security_ids:
        sec_id_strs = None
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])
        
    if sec_id_strs:
        query = f"""
        select SECODE as SECURITYID, FDNAME, SNAMECOMP, FSYMBOL, FDNATURE, INVESTSTYLE FROM TQ_FD_BASICINFO
        WHERE
            ISVALID = 1 AND
            SECODE in ({sec_id_strs}) AND
            (LIQUENDDATE >= '{trade_dt}' or LIQUENDDATE = '19000101')
        """
    else:
        query = f"""
        select SECODE as SECURITYID, FDNAME, SNAMECOMP, FSYMBOL, FDNATURE, INVESTSTYLE FROM TQ_FD_BASICINFO
        WHERE
            ISVALID = 1 AND
            (LIQUENDDATE >= '{trade_dt}' or LIQUENDDATE = '19000101')
        """
    return read_sql(query)


def fd_typeclass(security_ids, trade_dt=None):
    if not trade_dt:
        trade_dt = dt.datetime.today().strftime("%Y%m%d")
    else:
        trade_dt = trade_dt.replace("-", "")

    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])

    query = f"""
    select TQ_FD_BASICINFO.SECODE as SECURITYID,
           TQ_FD_TYPECLASS.L1CODE,
           TQ_FD_TYPECLASS.L1NAME,
           TQ_FD_TYPECLASS.L2CODE,
           TQ_FD_TYPECLASS.L2NAME,
           TQ_FD_TYPECLASS.L3CODE,
           TQ_FD_TYPECLASS.L3NAME from TQ_FD_TYPECLASS join TQ_FD_BASICINFO on TQ_FD_BASICINFO.SECURITYID = TQ_FD_TYPECLASS.SECURITYID
    WHERE
        TQ_FD_TYPECLASS.ISVALID = 1 AND
        TQ_FD_BASICINFO.ISVALID = 1 AND
        TQ_FD_BASICINFO.SECODE in ({sec_id_strs}) AND
        (TQ_FD_TYPECLASS.ENDDATE >= '{trade_dt}' or TQ_FD_TYPECLASS.ENDDATE = '19000101')
    """
    return read_sql(query)


def fd_hshkiport(security_ids, report_dates_begin, report_dates_end=None):

    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])
    
    if report_dates_end:
        query = f"""
        select SECODE as SECURITYID, INDCLASSCODE, INDUSTRYCODE, INDUSTRYNAME, REPORTDATE, MVALUE, ACCNETMKTCAP FROM TQ_FD_HSHKIPORT
        WHERE
            REPORTDATE >= '{report_dates_begin}' AND
            REPORTDATE <= '{report_dates_end}' AND
            ISVALID = 1 AND
            INDCLASSCODE = '2102' AND
            SECODE in ({sec_id_strs})
        """
    else:
        query = f"""
        select SECODE as SECURITYID, INDCLASSCODE, INDUSTRYCODE, INDUSTRYNAME, REPORTDATE, MVALUE, ACCNETMKTCAP FROM TQ_FD_HSHKIPORT
        WHERE
            REPORTDATE >= '{report_dates_begin}' AND
            ISVALID = 1 AND
            INDCLASSCODE = '2102' AND
            SECODE in ({sec_id_strs})
        """
    return read_sql(query).sort_values("SECURITYID").reset_index(drop=True)


def fd_assetportfolio(security_ids, report_dates_begin, report_dates_end=None):
    # 获取相关组合情况
    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])
    if report_dates_end:
        query = f"""
        SELECT SECODE as SECURITYID, REPORTDATE, BDRTO, CONVBDRTO, EQUITYINVERTO FROM TQ_FD_ASSETPORTFOLIO
        WHERE
            REPORTDATE >= '{report_dates_begin}' AND
            REPORTDATE <= '{report_dates_end}' AND
            ISVALID = 1 AND
            SECODE in ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT SECODE as SECURITYID, REPORTDATE, BDRTO, CONVBDRTO, EQUITYINVERTO FROM TQ_FD_ASSETPORTFOLIO
        WHERE
            REPORTDATE >= '{report_dates_begin}' AND
            ISVALID = 1 AND
            SECODE in ({sec_id_strs})
        """
    return read_sql(query).sort_values("SECURITYID").reset_index(drop=True)


def fd_qtfdnav(security_ids, trade_dt):
    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])

    query = f"""
    SELECT SECODE as SECURITYID, UNITNAV, UNITACCNAV, REPAIRUNITNAV, NAVGRTD FROM TQ_FD_DERIVEDN
    WHERE
        ENDDATE = '{trade_dt}' AND
        ISVALID = 1 AND
        SECODE in ({sec_id_strs})
    """
    return read_sql(query).sort_values("SECURITYID").reset_index(drop=True)


def _to_yyyy_mm_dd(trade_dt):
    if len(trade_dt) == 10:
        return trade_dt
    else:
        return f"{trade_dt[0:4]}-{trade_dt[4:6]}-{trade_dt[6:8]}"


def _security_ids(security_ids):
    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    elif security_ids is None:
        sec_id_strs = None
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])
    return sec_id_strs


def fd_skdetail(security_ids, trade_dt):
    p_id_strs = _security_ids(security_ids)

    query = f"""
    SELECT SECODE as pid, SKCODE as security_code, SKNAME, HOLDMKTCAP, HOLDAMT, NAVRTO FROM TQ_FD_SKDETAIL
    WHERE
        ENDDATE = '{trade_dt}' AND
        ISVALID = 1 AND
        SECODE in ({p_id_strs})
    """
    df = read_sql(query).sort_values("security_code")
    total_value = df.HOLDMKTCAP.sum()
    df["weight"] = df.HOLDMKTCAP / total_value
    return df.reset_index(drop=True)


def risk_exposure(start_date, end_date, security_ids=None):
    sec_id_strs = _security_ids(security_ids)

    if sec_id_strs:
        query = f"""
        SELECT trade_date, security_code, {','.join(RISK_STYLES)}, {','.join(INDUSTRY_STYLES)} FROM risk_exposure
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1 AND
            security_code IN ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT trade_date, security_code, {','.join(RISK_STYLES)}, {','.join(INDUSTRY_STYLES)} FROM risk_exposure
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def risk_return(start_date, end_date):
    query = f"""
            SELECT trade_date, {','.join(RISK_STYLES)}, {','.join(INDUSTRY_STYLES)} FROM risk_return
            WHERE 
                trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
                trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
                flag = 1
            """
    return read_sql(query, source="mysql").sort_values(["trade_date"]).reset_index(drop=True)


def specific_return(start_date, end_date, security_ids=None):
    sec_id_strs = _security_ids(security_ids)

    if sec_id_strs:
        query = f"""
        SELECT trade_date, security_code,  secShortName, symbol, spret FROM specific_return
        WHERE 
          trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
          trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
          flag = 1 AND
            security_code IN ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT trade_date, security_code,  secShortName, symbol, spret FROM specific_return
        WHERE 
          trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
          trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
          flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def risk_cov(start_date, end_date, model="short"):
    query = f"""
        SELECT trade_date, FactorID, {','.join(RISK_STYLES)}, {','.join(INDUSTRY_STYLES)} FROM risk_cov_{model}
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "FactorID"]).reset_index(drop=True)


def special_risk(start_date, end_date, model="short", security_ids=None):
    sec_id_strs = _security_ids(security_ids)

    if sec_id_strs:
        query = f"""
        SELECT trade_date, security_code, secShortName, symbol, SRISK FROM specific_risk_{model}
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1 AND
            security_code IN ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT trade_date, security_code, secShortName, symbol, SRISK FROM specific_risk_{model}
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def index_component(index_code, start_date, end_date):
    query = f"""
            SELECT index_component.trade_date, index_component.security_code, `index`.sname, `index`.weighing / 100.0 AS weight FROM `index`
            JOIN index_component ON
              `index`.trade_date = index_component.trade_date AND
              `index`.isymbol = index_component.isymbol AND
              `index`.symbol = index_component.symbol
            WHERE
              `index`.trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
              `index`.trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
              `index`.flag = 1 AND
              index_component.flag = 1 AND
              index_component.isecurity_code = '{index_code}';
              """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def sw_industry_daily(start_date, end_date, level=1, security_ids=None):
    sec_id_strs = _security_ids(security_ids)
    if sec_id_strs:
        query = f"""
        SELECT trade_date, security_code, industry_code{level} as industry_code, industry_name{level} as industry_name FROM sw_industry_daily
        WHERE
          trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
          trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
          flag = 1 AND
          security_code IN ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT trade_date, security_code, industry_code{level} as industry_code, industry_name{level} as industry_name FROM sw_industry_daily
        WHERE
          trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
          trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
          flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def stk_return(start_date, end_date, security_ids=None):
    sec_id_strs = _security_ids(security_ids)
    if sec_id_strs:
        query = f"""
        SELECT security_code, min(trade_date) as start_date, max(trade_date) as end_date, exp(sum(log(1.0 + change_pct / 100.))) AS chg_pct FROM stk_daily_price_pro
        WHERE
              trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
              trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
              security_code IN ({sec_id_strs}) AND
              flag = 1
        GROUP BY security_code;
        """
    else:
        query = f"""
        SELECT security_code, min(trade_date) as start_date, max(trade_date) as end_date, exp(sum(log(1.0 + change_pct / 100.))) - 1.0 AS chg_pct FROM stk_daily_price_pro
        WHERE
              trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
              trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
              flag = 1
        GROUP BY security_code;
        """
    return read_sql(query, source="mysql").sort_values("security_code").reset_index(drop=True)
